package ppt;

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.ArrayList;
import java.util.List;

import javax.imageio.ImageIO;

import net.contentobjects.jnotify.JNotify;
import net.contentobjects.jnotify.JNotifyException;
import net.contentobjects.jnotify.JNotifyListener;

import org.apache.log4j.Logger;

/**
 * The ImageServer class is a stand alone application which opens socket, 
 * listens for a new client connection. The client is added to a list when connected.
 * The server gets notified upon each PNG file creation in file system 
 * where Power Point Capture Show add-in stores slide as an image. 
 * The server streams the image file to currently connected clients. 
 *   
 * @author Tatyana Steinberg
 *
 */
public class ImageServer extends Thread implements JNotifyListener {
	
	static Logger logger = Logger.getLogger(ImageServer.class);
	
	// server port
	static int port;
	
	// list of connected clients 
	static List clients;
	
	BufferedImage currentImage;
	
	
	
	/**
	 * @param args: IMG_DIR and SENDER_PORT arguments can be sent
	 *   
	 */
	public static void main(String[] args) {
		
		if (args.length != 2) {
			System.out.println("Image dirctory and port must be specified. Exiting...");
		}
		
		String imgDir = args[0];
		port = Integer.parseInt(args[1]);
		
		// initialize server
		ImageServer server = new ImageServer();
		server.start();
		
		// we are interested only in file creation event
		// int mask = JNotify.FILE_CREATED  |  JNotify.FILE_DELETED  |  JNotify.FILE_MODIFIED  |  JNotify.FILE_RENAMED; 
		int mask = JNotify.FILE_CREATED;
		
		//	watch subtree? 
		boolean watchSubtree = false; 

		// add actual watch 
		try {
			JNotify.addWatch(imgDir, mask, watchSubtree, server);
		} catch (JNotifyException e) {
			logger.error(e,e);
		} 

	}
	
	
	
	/** 
	 * @see java.lang.Thread#run()
	 * The metnod initializes client list, opens socket, and listens for 
	 * a new client connection 
	 */
	public void run()
	{
		// initialize client list
		clients = new ArrayList();
		
		ServerSocket srvr=null;
		try	{
			srvr = new ServerSocket(port);
		} catch (Exception e) {
			logger.error("Socket Exception", e);
			System.exit(0);
		}
		while(true) {
			
			try	{
			
				Socket socket = srvr.accept();
				socket.setTcpNoDelay(true);
				socket.setKeepAlive(true);
				logger.debug("NEW CONNECTION " + socket);

				clients.add(new ImageStreamer(socket.getOutputStream(), currentImage));
			}
			catch (Exception e)	{
				logger.error("Socket Exception", e);
			}
		}


	}
	
	

	
	/** 
	 * The method reads newly created file from file system
	 * and assigns it to list of clients - ImageStreamer objects
	 * @see net.contentobjects.jnotify.JNotifyListener#fileCreated(int, java.lang.String, java.lang.String)
	 * 
	 */
	public void fileCreated(int wd, String rootPath, String name) {

		logger.info("New file: " + rootPath + "/" + name);

		// got new image, now create image object and distribute it to clients
		BufferedImage bi = null;
		int timeoutCount = 0;
		
		while(bi == null && timeoutCount <= 30) {
			try {
				timeoutCount++;
				bi = ImageIO.read(new File(rootPath + "/" + name));
			} catch (IOException e) {
				logger.warn("File " + name + " is not available, waiting...");
				bi = null;
			}
		}
		
		// if image still null
		if (bi == null) {
			logger.error("Unable to read file " + rootPath + "/" + name);
			return;
		}
		
		logger.info("Got new image: " + rootPath + "/" + name);

		// save it in case someone join presentation later
		currentImage = bi;
		
		for (int i=0; i < clients.size(); i++) {
			ImageStreamer streamer = (ImageStreamer) clients.get(i);
			streamer.setImage(bi);
		}
	
	}

	
	
	/* (non-Javadoc)
	 * @see net.contentobjects.jnotify.JNotifyListener#fileDeleted(int, java.lang.String, java.lang.String)
	 */
	public void fileDeleted(int wd, String rootPath, String name) {}

	
	/* (non-Javadoc)
	 * @see net.contentobjects.jnotify.JNotifyListener#fileModified(int, java.lang.String, java.lang.String)
	 */
	public void fileModified(int wd, String rootPath, String name) {}

	
	/* (non-Javadoc)
	 * @see net.contentobjects.jnotify.JNotifyListener#fileRenamed(int, java.lang.String, java.lang.String, java.lang.String)
	 */
	public void fileRenamed(int wd, String rootPath, String oldName, String newName) {}

	
}
