package ppt;

import java.awt.image.BufferedImage;
import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import javax.imageio.ImageIO;

import org.apache.log4j.Logger;

/**
 * The ImageStreamer class is used as client in ImageServer.
 * The class main functionality is to receive PNG file as BufferedImage, 
 * divide the image by squares. The square then get compared with the previous ones 
 * from the same location. The new square gets sent to 
 * the client via real time protocol - RTP
 *  
 * @author Tatyana Steinberg
 *
 */
public class ImageStreamer {

	static Logger logger = Logger.getLogger(ImageStreamer.class);
	
	// socket output stream
	OutputStream out;
	
	// initialize flag
	boolean initialized = false;

	// curent image
	BufferedImage image;
	int imageWidth = 100; // 100 is default to open initial window if there is no image yet
	int imageHeight = 100;
	
	// height and width of subimages
	static int tileWidth = 32;
	static int tileHeight = 32;

	// array to keep image fragments
	Object[] tiles = null;

	// number of columns and rows in the grid
	int numOfCols;
	int numOfRows;

	RTPStream rs;
	

	/**
	 * Constructor which is used for each client to open a window, initialize 
	 * an array for storing squares with dimensions according to the first image 
	 * recived.
	 * 
	 * @param socketOutputStream as OutputStream
	 * @param bufImage as BufferedImage
	 */
	public ImageStreamer(OutputStream socketOutputStream, BufferedImage bufImage) {
		out = new BufferedOutputStream(socketOutputStream);
		image = bufImage;
		
		try {
			if (image == null) {
				openClientWindow(true);
			}
			else {
				initialize();
				openClientWindow(false);
				sendImage();
			}

		} catch (IOException e) {
				logger.error(e);
		}
	}


	/**
	 * The method sets image for every new png created file  
	 * and sends the image to client
	 * 
	 * @param bufImage as BufferedImage
	 */
	public void setImage(BufferedImage bufImage) {
		
		this.image = bufImage;
		
		imageWidth = image.getWidth();
		imageHeight = image.getHeight();
		
		try {		
			// if this is the first image
			if (! initialized) {
				initialize();
				// resize window - packet two
				out.write(resizeWindowPacket());
			}

			sendImage();
		
		} catch (IOException e) {
			logger.error(e);
		}
		
	}


	
	/**
	 * The method initializes array of subimages-tiles by dimensions
	 *  The dimensions depend on the square width and height 
	 */
	private void initialize() {
		
		// get image width and height
		imageWidth = image.getWidth();
		imageHeight = image.getHeight();

		// get nomber of cols / rowa
		numOfCols = imageWidth / tileWidth;

		if (imageWidth % tileWidth > 0)
			numOfCols++;

		numOfRows = imageHeight / tileHeight;

		if (imageHeight % tileHeight > 0)
			numOfRows++;

		// allocate array
		tiles = new Object[numOfCols * numOfRows];

		initialized = true;

	}

	
	/**
	 * The method opens client window, creates and sends RTP packets
	 * with code 0 - open window
	 * code 1 - close window (currently is not used by the Sender)
	 * code 2 - resize window
	 * code 3 - send data
	 * 
	 * @param withBanner as boolean. 
	 * In case of now current PNG images to send default banner.png is shown
	 * 
	 * @throws IOException 
	 * 
	 */
	private void openClientWindow(boolean withBanner) throws IOException {

		BufferedImage bi = null;
		
		if (withBanner) {
			// load and display initial banner
			InputStream is = getClass().getClassLoader().getResourceAsStream("banner.PNG");
			
			bi = ImageIO.read(is);

			imageWidth = bi.getWidth();
			imageHeight = bi.getHeight();
		}
		// open window - packet one
		out.write(openWindowPacket());

		// resize window - packet two
		out.write(resizeWindowPacket());

		if (withBanner) {
			// send initial banner
			out.write(dataPacket(imageToByteArray(bi), 0, 0, imageWidth,
					imageHeight, true));
		}
		
		out.flush();

	}


	
	/**
	 * The method gets splits image on squares. A data packet is created and gets sent
	 * for each new square after comparison with the previous one 
	 * from the same location if such existed. 
	 *  
	 * @throws IOException
	 */
	public void sendImage() throws IOException {

		int tilePointer = 0;

		int x = 0;
		int y = 0;
		int w = 32;
		int h = 32;

		// process image / split into cells

		// rowHeight equals tileHeight except for the last row,
		// same for col width
		h = tileHeight;

		// loop thru rows
		for (int r = 0; r < numOfRows; r++) {

			x = 0; // reinitialize row position
			w = tileWidth; // reinitialize cell width

			// for last row height is smaller
			if (r == numOfRows - 1)
				h = imageHeight % tileHeight;
			
			if(h == 0  || y == imageHeight)
				break;

			// loop thru columns
			for (int c = 0; c < numOfCols; c++) {
				// for last column width is smaller
				if (c == numOfCols - 1)
					w = imageWidth % tileWidth;

				if(w == 0 || x == imageWidth)
					break;
				// now get tile from the image
				// int currTile[] = bi.getRGB(x, y, w, h, new int[w*h], 0, w);
				byte[] currTile = imageToByteArray(image.getSubimage(x, y, w, h));

				// get prev tile to compare
				byte[] prevTile = (byte[]) tiles[tilePointer];

				if (prevTile == null || !equals(currTile, prevTile)) {

					byte[] packet = dataPacket(currTile, x, y, w, h, true);
					
					out.write(packet);
					//logger.debug("Sent tile " + tilePointer);
					
					// create a packet and send tile
					tiles[tilePointer] = currTile;
				} // if prev not = curr

				tilePointer++;
				x += w;
			} // for cols

			y += h;
		} // for rows
		out.flush();
	}
	
	
	
	
	/**
	 * The method opens a window (code 0)
	 * @return byte array 
	 */
	public byte[] openWindowPacket() {

		byte[] b = new byte[22];
		int length =  ( 12 + 8 );
		int code = 0;
		int id = 1;

		rs = new RTPStream(101);
		b = createPacketHeader(code,id, length, false);

		return b;
	}

	
	
	/**
	 * The method resizes window (code 2)
	 * @return byte array 
	 */
	public byte[] resizeWindowPacket() {
		
		byte[] b = new byte[38];
		int length =  ( 12 + 8 +16 );
		int code = 2;
		int id = 1;
		int x = 22;

		b = createPacketHeader(code,id, length, false);
		b = createWindowDimensions(b, 0, 0, imageWidth, imageHeight, x);

		return b;
	}

	
	
	
	/**
	 * THe method creates packet with data. (code 3)
	 * @param buffer contained subimage
	 * @param l - left position for the subimage 
	 * @param t - top position for the subimage
	 * @param w - width of of subimage
	 * @param h - height of the subimage
	 * @param isMarker is true for each packet
	 * 
	 * @return byte array contained header, dimensions and subimage in bytes 
	 * before encoding it on client side
	 */
	public byte[] dataPacket(byte[] buffer, int l, int t, int w, int h, boolean isMarker) {

		byte[] b = new byte[38 + buffer.length];
		int length = (12 + 8 + 16 + buffer.length);

		int code = 3;
		int id = 1;
		int x = 22;

		b = createPacketHeader(code, id, length, isMarker);
		b = createWindowDimensions(b, l, t, w, h, x);

		x = 38;
		for (int i = 0; i < buffer.length; i++) {
			b[x + i] = buffer[i];
		}

		return b;

	}

	
	
	
	
	/**
	 * The method creates header as byte array 
	 * 
	 * @param code
	 * @param id
	 * @param length
	 * @param isMarker
	 * @return byte array for a header, code and window id
	 */
	public byte[] createPacketHeader(int code, int id, int length, boolean isMarker){

		byte[] b = new byte[length+2];
		//int length =  ( 12 + 8 );

		int x = 14;

		//RTPStream rs = new RTPStream(101);
		rs.isMarker(isMarker);
		byte[] header = rs.getRTPHeader(0);

		b[0]=(byte)(0xFF & (length>>8));
		b[1]=(byte)(0xFF & (length));

	///	logger.debug("tester buff "+(int)(((b[0] & 0xff) << 8) | (b[1] & 0xff)));
		System.arraycopy(header, 0, b, 2, 12);

		b[x++] =(byte)(0xFF & (code>>24));
		b[x++] =(byte)(0xFF & (code>>16));
		b[x++] =(byte)(0xFF & (code>>8));
		b[x++] = (byte)(0xFF & (code));
		b[x++] =(byte)(0xFF & (id>>24));
		b[x++] =(byte)(0xFF & (id>>16));
		b[x++] =(byte)(0xFF & (id>>8));
		b[x++] =(byte)(0xFF & (id));

		return b;
	}

	
	
	/**
	 * The method adds creates dimensions for window in byte array
	 * It is used in packets for resizing a window (code 2) 
	 * 
	 * @param b - byte array
	 * @param left
	 * @param top
	 * @param width
	 * @param heigth
	 * @param x - position in byte array which starts after header end position
	 * @return byte array for left, top, width and heigth window sizes
	 */
	public byte[] createWindowDimensions(byte[] b, int left, int top, int width, int heigth, int x){

		b[x++] =(byte)(0xFF & (left>>24));
		b[x++] =(byte)(0xFF & (left>>16));
		b[x++] =(byte)(0xFF & (left>>8));
		b[x++] =(byte)(0xFF & (left));
		b[x++] =(byte)(0xFF & (top>>24));
		b[x++] =(byte)(0xFF & (top>>16));
		b[x++] =(byte)(0xFF & (top>>8));
		b[x++] =(byte)(0xFF & (top));

		b[x++] =(byte)(0xFF & (width>>24));
		b[x++] =(byte)(0xFF & (width>>16));
		b[x++] =(byte)(0xFF & (width>>8));
		b[x++] =(byte)(0xFF & (width));
		b[x++] =(byte)(0xFF & (heigth>>24));
		b[x++] =(byte)(0xFF & (heigth>>16));
		b[x++] =(byte)(0xFF & (heigth>>8));
		b[x++] =(byte)(0xFF & (heigth));

		return b;
	}


	

	/**
	 * The method gets BufferedImage as byte array
	 * 
	 * @param bi as BufferedImage
	 * @return byte array
	 */
	public byte[] imageToByteArray(BufferedImage bi) {
	
		ByteArrayOutputStream baos = new ByteArrayOutputStream(10000);
		try {
			ImageIO.write(bi, "PNG", baos);
		} catch (IOException e) {
			logger.error(e);
		}
		return baos.toByteArray();
	}
	
	
	
	/**
	 * The method compares two subimages as byte arrays
	 * @param array1 as byte array
	 * @param array2 as byte array
	 * @return boolean
	 */
	public boolean equals(byte[] array1, byte[] array2) {

		 if (array1.length != array2.length)
		 	return false;

		for (int i=0; i < array1.length; i+=4) {

			if (array1[i] != array2[i]) {
				return false;
			}
		}

		return true;
	}
	
	
	
}
